/*
Public Key Crypto for UDP Encryption
by Joel R. Voss
Feb 2, 2006

Looking at the OpenSSH source, we can see some cool stuff.

If I want to offload the DSA keygen stuff, I can use the utility ssh-keygen. It 
is easy as heck.

ssh-keygen -t dsa -f dsa1.key

The public key is in an easy file dsa1.key.pub

Then to load the private key, I have to do something pretty simple.

There are times when I have more than one key pair.
At that point, I want to use public keys to see if I have a match.
When I find a match, I'll stop. But really, the public key part is too easy:
read file, b64decode key, memcmp.
It's not really a pubkey at all.
The only reason why we'd want to load the pubkey from the file is if it takes a 
bunch of work to get the pubkey from the privkey.
key_demote()

DSA can only sign, so wtf do I do with it? I can authenticate easily.
DSA has no key exchange.

DH! Diffie Hellman provides a very easy access to 256 bytes of secret data with 
the simple exchange of public keys + one-way encrypted secret.

My understanding of DH (which might be incorrect) is that DH generates 
pseudo-random keypair for each session which is known only by the each party. 
The public keys are exchanged. With A's private and B's public you can 
generate the secret. With B's private and A's public you can generate the 
secret.

*/

#include <openssl/dsa.h>
#include <openssl/engine.h>

//#include <openssl/bn.h>

#include <openssl/evp.h>
#include <openssl/pem.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "keyex1.h"

//void fatal(const char *data);

DSA *key_new()
{
    DSA *dsa;
        if ((dsa = DSA_new()) == NULL)
            return 0; //fatal("key_new: DSA_new failed");
        if ((dsa->p = BN_new()) == NULL)
            return 0; //fatal("key_new: BN_new failed");
        if ((dsa->q = BN_new()) == NULL)
            return 0; //fatal("key_new: BN_new failed");
        if ((dsa->g = BN_new()) == NULL)
            return 0; //fatal("key_new: BN_new failed");
        if ((dsa->pub_key = BN_new()) == NULL)
            return 0; //fatal("key_new: BN_new failed");
    return dsa;
}

void key_free(DSA *dsa)
{
    if (dsa != NULL)
        DSA_free(dsa);
    dsa = NULL;
}

DSA *key_load_private_pem(const char *filename, const char *passphrase)
{
    FILE *fp;
    EVP_PKEY *pk = NULL;
    DSA *prv = NULL;
    int fd;
    
    OpenSSL_add_all_algorithms();
    
    // The following line causes a segfault. Why? Doh.
    //fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0600);
    fd = open(filename, O_RDONLY);
    fp = fdopen(fd, "r");
    
    pk = PEM_read_PrivateKey(fp, NULL, NULL, (char *)passphrase);
    
    prv = key_new();
    prv = EVP_PKEY_get1_DSA(pk);
    
    fclose(fp);
    if (pk != NULL)
        EVP_PKEY_free(pk);
    return prv;
}

/* Converts a private to a public key */
DSA *key_demote(const DSA *k)
{
    DSA *pk;

    if ((pk = DSA_new()) == NULL)
        return 0; //fatal("key_demote: DSA_new failed");
    if ((pk->p = BN_dup(k->p)) == NULL)
        return 0; //fatal("key_demote: BN_dup failed");
    if ((pk->q = BN_dup(k->q)) == NULL)
        return 0; //fatal("key_demote: BN_dup failed");
    if ((pk->g = BN_dup(k->g)) == NULL)
        return 0; //fatal("key_demote: BN_dup failed");
    if ((pk->pub_key = BN_dup(k->pub_key)) == NULL)
        return 0; //fatal("key_demote: BN_dup failed");
    return (pk);
}

BIGNUM *get_my_public_key(DSA *dsa)
{
    return dsa->pub_key;
}

BIGNUM *recv_dsa_public_key()
{
    BIGNUM *pub_key = BN_new();
    return pub_key;
}

/*
Diffie Hellman Parameters generated with 
openssl dsaparam 2048 -out dsap1.key
openssl dhparam -dsaparam -C -in dsap1.key
*/
DH *get_dh2048()
{
    static unsigned char dh2048_p[]={
        0x93,0x95,0x06,0xC1,0x39,0x30,0x2F,0xEE,0xB5,0x50,0x9D,0x36,
        0x18,0x1C,0x3D,0x9D,0xC4,0xDF,0x92,0x9C,0xDE,0xD1,0x48,0xF1,
        0x50,0x66,0x6B,0xA6,0x09,0x51,0xDF,0x8C,0xB4,0x1C,0x4C,0x27,
        0x1C,0xDE,0x6B,0xEF,0x7D,0x71,0x36,0x7B,0xC8,0x62,0x9F,0x68,
        0x0B,0x6E,0xEB,0x39,0x6A,0x84,0xD9,0xE0,0x09,0x52,0xE6,0x08,
        0x77,0x58,0x1B,0xDA,0xEA,0xED,0x87,0xAE,0xF1,0x22,0x53,0x23,
        0x05,0x5B,0xAE,0xC8,0xD2,0xFE,0x0E,0xBC,0x0B,0x66,0x5F,0xC5,
        0xD3,0xC3,0xAD,0x96,0x51,0x18,0x2E,0xA9,0xE1,0x5D,0x48,0x93,
        0x18,0x08,0x64,0xDD,0xA2,0x9A,0xC4,0x4D,0x0E,0x11,0xE8,0xBF,
        0x80,0xED,0xD2,0xE5,0x46,0x79,0xCB,0xF1,0xD7,0x30,0x9C,0x1C,
        0xA0,0xC1,0x25,0xEE,0x73,0xDD,0x07,0x82,0xCC,0x35,0x68,0xC1,
        0x2A,0x21,0x4E,0xCE,0xBF,0xAC,0x45,0x00,0xAE,0x92,0x5B,0x51,
        0x93,0x29,0x56,0x49,0xB8,0x95,0x5A,0x51,0x1D,0xEC,0xFC,0xCB,
        0x3F,0x0B,0x59,0x9B,0x32,0xC0,0x0F,0x7B,0xBA,0x6C,0x20,0xD0,
        0x91,0x67,0xB0,0xA2,0xD9,0x3A,0xB4,0x51,0x70,0x8F,0xB7,0xE6,
        0xA6,0xB3,0x88,0xEC,0x1A,0x0F,0xAE,0xBE,0xEE,0x58,0xE8,0x87,
        0x6B,0x1C,0xEB,0xE5,0xD0,0x14,0x95,0x02,0xCA,0x35,0xB0,0xD0,
        0xE1,0x33,0x3D,0x5A,0x12,0x36,0xF7,0xF0,0x80,0x09,0xC4,0xBD,
        0x11,0x1D,0xD7,0x4A,0x57,0xE4,0xCF,0x84,0x3C,0xB5,0x53,0x9A,
        0x86,0xA7,0x34,0x4C,0x40,0x0D,0x86,0xDD,0x14,0x9A,0xE5,0x10,
        0xBB,0xD4,0xE7,0x0B,0x9E,0x98,0xE7,0x87,0x8A,0x05,0xE1,0x88,
        0xE6,0x8A,0xE2,0x91,
        };
    static unsigned char dh2048_g[]={
        0x34,0xA2,0xB5,0x2E,0xAD,0x65,0x47,0x70,0x00,0x8F,0x7B,0xA7,
        0x20,0x57,0xC6,0xC9,0x8C,0x32,0x23,0x96,0x6C,0x5D,0x2D,0x34,
        0x86,0xB2,0x38,0xCA,0x51,0x94,0x41,0x0C,0x64,0xE9,0xDB,0x2F,
        0x84,0x26,0x62,0x6C,0xFA,0x5D,0xD9,0x16,0xAB,0x29,0x0D,0xE5,
        0x6F,0x6F,0xBD,0xEE,0x3C,0x40,0xED,0xEE,0x26,0x2C,0x67,0x45,
        0x55,0x3D,0x33,0x2F,0x62,0x1A,0x67,0x91,0x07,0x6A,0x42,0x52,
        0x1D,0xFD,0x80,0x60,0xB5,0xF8,0x4B,0x2A,0x45,0xF2,0xAC,0x3F,
        0x45,0xDE,0xAC,0xC2,0x0D,0xC6,0xA7,0x40,0xBD,0x2C,0xDE,0xA6,
        0xB7,0xCF,0xAA,0x74,0xDD,0xB4,0x95,0x38,0x1C,0xC2,0xD1,0x9F,
        0x82,0xE5,0x67,0x31,0xFE,0x23,0x5A,0xC5,0xAC,0x56,0x61,0x0C,
        0xF2,0xFF,0x04,0xEF,0x5F,0x87,0x69,0x4E,0xEB,0x53,0xA7,0xDA,
        0xB9,0x7E,0xD3,0xC4,0x69,0x6B,0x4D,0x77,0x0B,0x19,0x33,0x39,
        0xAA,0x75,0x38,0x3F,0x86,0x8C,0xC7,0xB3,0xAE,0xDE,0x26,0x87,
        0x1F,0x41,0xA0,0xDB,0x60,0x31,0x5F,0xDD,0x15,0x2B,0x22,0x1D,
        0xFB,0xBD,0x65,0xEA,0x11,0xE8,0x92,0xD0,0xF8,0xC0,0x7D,0xA5,
        0xBB,0x3A,0x52,0x3B,0x96,0x95,0x67,0x14,0x33,0xEE,0xBA,0x80,
        0x43,0xD5,0x32,0x2B,0x78,0x6E,0xDF,0x0E,0xD9,0x50,0xAC,0xF8,
        0xE9,0x06,0x23,0xD3,0x9C,0xE3,0x1D,0xFA,0xF3,0xE1,0x66,0x2A,
        0xA0,0x2A,0xC9,0x7C,0x3E,0xAE,0x9D,0x92,0x1E,0x95,0x35,0x01,
        0x7F,0x23,0x03,0xE1,0x92,0xFE,0xA4,0xC2,0xCD,0x1B,0x65,0x19,
        0x8A,0x1B,0x2C,0xA0,0x2F,0x0E,0xCC,0xA9,0xEB,0xAA,0x9D,0xDC,
        0x81,0xE3,0xDB,0xD4,
        };
    DH *dh;
    
    if ((dh=DH_new()) == NULL) return(NULL);
    dh->p=BN_bin2bn(dh2048_p,sizeof(dh2048_p),NULL);
    dh->g=BN_bin2bn(dh2048_g,sizeof(dh2048_g),NULL);
    if ((dh->p == NULL) || (dh->g == NULL))
            { DH_free(dh); return(NULL); }
    dh->length = 160;
    return(dh);
}

int dh_gen_key_pair(BIGNUM **dh_pub_key, DH **dh_out)
{
	int ret;
	DH *dh;
    // p and g are calculated in design time.
    dh = get_dh2048();
    
    // Generate the private key.
	if (!BN_rand(dh->priv_key, 2*1024, 0, 0))
    {
        printf("dh_gen_key: BN_rand failed");
        return -1;
    }
	ret = DH_generate_key(dh);
    
    //ret = DH_check(dh, &codes);
	*dh_pub_key = BN_dup(dh->pub_key);
    *dh_out = dh;
    return ret;
}

int dh_gen_secret(DH *dh, BIGNUM *other_dh_pub_key, unsigned char *secret_out, int *secret_len)
{
    
    int size = DH_size(dh);
	// Decide whether this is an error, a warning, or nothing.
	if(size > *secret_len)
	{
		// This is an error condition. OpenSSL wants more than we can give it.
		return 1;
	}
	else if(size < *secret_len)
	{
		// This is a warning. OpenSSL wants less than we have available.
		*secret_len = size;
	}
    // Secret size is 256 bytes, plenty.
    //printf("Secret size: %i;\n", size);
    //*secret_out = malloc(size);
    
    int kout = DH_compute_key(secret_out, other_dh_pub_key, dh);

	//BIGNUM *shared_secret = BN_new();
	//BN_bin2bn(*secret_out, kout, shared_secret);
    DH_free(dh);

    return 0;
}

#ifdef TEST

/*void fatal(const char *data)
{
    printf(data);
}

void print_hex(const unsigned char *data, int b)
{
    int i;

    for (i = 0;  i < b - 1;  i++)
        fprintf(stderr, "%02hhX:", data[i]);
    fprintf(stderr, "%02hhX\n", data[b - 1]);
}*/

BIGNUM *recv_dh_public_key(BIGNUM *other_dh_pub_key)
{
    BIGNUM *pub_key;
    DH *dh;
    dh_gen_key_pair(&pub_key, &dh);
    
    DH_free(dh);
    
    return pub_key;
}

void send_public_key(BIGNUM *pub_key)
{
    return;
}

#include <string.h>

#ifdef _OPENBSD
#define RANDOM_DEVICE "/dev/srandom"
#else
#define RANDOM_DEVICE "/dev/random"
#endif

int seed_prng(int bytes)
{
    // This slows you down if you use it very much.
    if (!RAND_load_file(RANDOM_DEVICE, bytes))
        return 0;
    return 1;
}

int main(int argc, char **argv)
{
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }
	
	BIO *out;
	out = BIO_new_fd(fileno(stdout), BIO_NOCLOSE);
    
    // Authentication with DSA.
	// FIXME: The key should be encrypted with a password given by the user.
    char password[256];
    strcpy(password, "testdsa");
    DSA *dsa_key = key_load_private_pem("dsa1.key", password);
    
    BIGNUM *dsa_pub_key = get_my_public_key(dsa_key);
	BIO_printf(out, "my public key: ");
	BN_print(out, dsa_pub_key);
    
	// FIXME: There is a requestor and a responder in the system.
	// The requestor sends first. The responder receives first.
    // Actually, if they both send first, it works. They do need to do a /poke. 
    // /poke also needs to be dropped by each client.

	// TODO: Send my Public Key.
    send_public_key(dsa_pub_key);
    // TODO: Receive their Public Key.
    BIGNUM *other_dsa_pub_key = recv_dsa_public_key();
	BIO_printf(out, "\ntheir public key: ");
	BN_print(out, other_dsa_pub_key);
	BIO_printf(out, "\n");
	
	// TODO: Check to ensure that their public key has been signed by my 
	// trusted group.
    
    // Secret Key Exchange using Diffie-Hellman
	
	BIGNUM *dh_pub_key = 0;
    DH *dh;
    // Generate a key pair
    int ret = dh_gen_key_pair(&dh_pub_key, &dh);
    // Send the public key.
    send_public_key(dh_pub_key);
    // Receive the peer's public key.
    BIGNUM *other_dh_pub_key = recv_dh_public_key(dh_pub_key);
    unsigned char secret[256];
	int secret_len = 256;
    // Create the secret key with this data.
    ret = dh_gen_secret(dh, other_dh_pub_key, secret, &secret_len);
    
    BIO_printf(out, "\nmy public key: ");
    BN_print(out, dh_pub_key);
    BIO_printf(out, "\n");
    BIO_printf(out, "\ntheir public key: ");
    BN_print(out, other_dh_pub_key);
    BIO_printf(out, "\n");
    
    BN_free(other_dh_pub_key);
    BN_free(dh_pub_key);
    // Do NOT free dsa_pub_key. It is part of dsa_key.
    BN_free(other_dsa_pub_key);
    key_free(dsa_key);

	BIO_free(out);
    
    free(secret);
    
    return 0;
}

#endif // TEST
